/*++

Copyright (c) 2004-2005 Alexander A. Telyatnikov (Alter)

Module Name:
    DbgPrintLog.cpp

Abstract:

Authors:
    Alexander A. Telyatnikov (Alter)
    + thanks to Max Sivkov for Service Mode

Environment:
    User mode only

Notes:

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Revision History:

--*/

#include <windows.h>
#include <stdio.h>
#include "DbgPrnHk.h"
#include "SvcManLib.h"
#include "Privilege.h"
#include "flusher.h"
#include "DbgPrintLog.h"
#include "dbgprint_opt.h"
#include "..\sdk\PostDbgMesg.h"
#include "fmt_output.h"
#include <Sddl.h>

//#include "dbgprn_options.h"

#define REG_CPU_DESCR_KEY_NAME_W  L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\%u"
#define REG_CPU_MHZ_VAL_NAME_W    L"~MHz"
#define REG_CPU_ID_VAL_NAME_W     L"Identifier"
#define REG_CPU_VENDOR_VAL_NAME_W L"VendorIdentifier"

void
WINAPI
service_main(
    DWORD dwArgc,
    LPTSTR *lpszArgv);

VOID AddToMessageLog(LPTSTR lpszMsg);

BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint);

SERVICE_TABLE_ENTRY dispatchTable[] =
{
    { NT_DbgPrnHk_Client_SVC_NAME, (LPSERVICE_MAIN_FUNCTION)service_main },
    { NULL, NULL }
};

WCHAR IdxFileN[MAX_PATH*2];
WCHAR LogFileN[MAX_PATH*2];
WCHAR LogFileSuf[MAX_PATH];

CHAR PathToExec[MAX_PATH] = ".";

BOOLEAN esc_key       = FALSE;
BOOLEAN pause_key     = FALSE;
BOOLEAN next_log      = FALSE;
BOOLEAN flush_file    = FALSE;
BOOLEAN sys_account   = FALSE;

DBGDUMP_OPT_RAW g_opt;
#define pOpt     (&g_opt)

/*
WCHAR  WDir[MAX_PATH*2];
PWCHAR LogFile = L"DbgPrint.log";  // default log-file name

BOOLEAN log_pid       = FALSE;
BOOLEAN log_tid       = FALSE;
BOOLEAN log_mode      = FALSE;
BOOLEAN log_irql      = FALSE;
BOOLEAN log_cpu       = FALSE;
//BOOLEAN log_time_tick = FALSE;

BOOLEAN log_time_ext  = TRUE;
BOOLEAN log_time_utc  = FALSE;
BOOLEAN log_time_perf = FALSE;

BOOLEAN log_time_date = FALSE;
BOOLEAN log_time_time = TRUE;
BOOLEAN log_time_nano = FALSE;

BOOLEAN skip_kernel   = FALSE;
BOOLEAN skip_umode    = FALSE;

//BOOLEAN stdout_mode   = FALSE;
//BOOLEAN stdout_copy   = FALSE;

BOOLEAN output_stdout = FALSE;
BOOLEAN output_file = TRUE;

//BOOLEAN change_config = FALSE;

ULONG FlushTimeout = 5;
BOOLEAN sync_mode  = FALSE;

ULONG MaxLogSize = 1024*1024*128; //128Mb
BOOLEAN prealloc_mode = FALSE;
BOOLEAN use_index       = FALSE;
ULONG MaxLogs = -1;
ULONG StartLogNum = 0;

BOOLEAN install_svc   = FALSE;
BOOLEAN install_drv   = FALSE;
BOOLEAN drv_very_first= FALSE;
BOOLEAN restart_drv   = FALSE;

ULONG   drv_mode      = SERVICE_DEMAND_START;
ULONG   svc_mode      = SERVICE_AUTO_START;

BOOLEAN deinstall     = FALSE;

BOOLEAN drv_opt       = FALSE;
*/

PDbgPrnHk_GetMessages_USER_OUT MsgBuffer = NULL;

// dbg messages capture buffers for user-mode
// (via OutputDebugString)
// '2' suffix  - buffer/counter for reading from get_messages()
// '2a' suffix - buffer/counter for writing from get_ODS_messages()
HANDLE hODS_lock = NULL;
ULONG MsgCount2 = 0;
ULONG MsgCount2a = 0;
PDbgPrnHk_GetMessages_USER_OUT MsgBuffer2 = NULL;
PDbgPrnHk_GetMessages_USER_OUT MsgBuffer2a = NULL;

ULONG BufferSize = 1024*1024;

ULONG LineNumber;

HANDLE StdIn  = NULL;
HANDLE StdOut = NULL;

HANDLE hODSreadyEvent = NULL;
HANDLE hDRVreadyEvent = NULL;

SERVICE_STATUS          ssStatus;       // current status of the service
SERVICE_STATUS_HANDLE   sshStatusHandle;

WCHAR UserName[MAX_PATH];

static const char help_text2[] = 
    "Commands:\n"
    "\t'Esc'   - exit\n"
    "\t'N'     - start new log\n"
    "\t'F'     - flush log buffer\n"
    "\t'H'     - display this help message\n"
    "Mode switchers (toggle on'off):\n"
    "\t'Space' - pause\n"
    "\t'S'     - synchronous mode\n"
    "\t'C'     - copy log to stdout\n"
    "\t'K'     - capture kernel messages\n"
    "\t'U'     - capture user-mode messages\n"
    ;

BOOLEAN
IsSystemAccount()
{
/*
    WCHAR UserName[MAX_PATH];
    ULONG len = MAX_PATH-1;

    if(!GetUserNameW(UserName, &len)) {
        UserName[0] = 0;
        return FALSE;
    }
    UserName[len] = 0;
    if(!wcsicmp(UserName, L"System")) {
        return TRUE;
    }
    return FALSE;
*/

    HANDLE              hToken;
    PTOKEN_USER         tu;
    DWORD               ReturnLength;
//    DWORD               i, j;
//    PCHAR               StrSid;
    PUCHAR              SubAuthCount;
    PDWORD              pRID;

    // obtain the token, first check the thread and then the process
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken)) {
        return FALSE;
    }

    GetTokenInformation(hToken, TokenUser, NULL, 0L, &ReturnLength);
    tu = (PTOKEN_USER)GlobalAlloc(GMEM_FIXED, ReturnLength);
    if(!GetTokenInformation(hToken, TokenUser, tu, ReturnLength, &ReturnLength)) {
        return FALSE;
    }
/*
    if(!ConvertSidToStringSid(tu->User.Sid, &StrSid)) {
        return FALSE;
    }
*/
    SubAuthCount = GetSidSubAuthorityCount(tu->User.Sid);
    if((*SubAuthCount) != 1)
        return FALSE;

    pRID = GetSidSubAuthority(tu->User.Sid, 0);
    if((*pRID) != SECURITY_LOCAL_SYSTEM_RID)
        return FALSE;
/*
    j = (*SubAuthCount);
    for(i=0; i<j; i++) {
        pRID = GetSidSubAuthority(tu->User.Sid, i);

        printf("i=%x, RID=%x\n", i, *pRID);
        sprintf((PCHAR)UserName, "i=%x, RID=%x\n", i, *pRID);
        OutputDebugString((PCHAR)UserName);
        continue;

        if((*pRID) != SECURITY_LOGON_IDS_RID ||
            i != 0) {
            return FALSE;
        }
        if((*pRID) != SECURITY_LOCAL_SYSTEM_RID ||
            i != 1) {
            return FALSE;
        }
    }

    Sleep(10000);
    ExitProcess(0);
*/

    return TRUE;

} // end IsSystemAccount()

VOID
TeeOutput(
    HANDLE h,
    PCHAR Buffer,
    ULONG Length
    )
{
    if(g_opt.output_file) {
        WriteFileBuffer(h, Buffer, Length);
    }
    if(g_opt.output_stdout) {
        WriteConsole(StdOut, Buffer, Length, &Length, NULL);
    }
} // end TeeOutput()

#define TeeOutputConst(h, str)  TeeOutput(h, str, sizeof(str)-1)

DBG_PRINT_LOG_IDX_HEADER IdxHeader;
DBG_PRINT_LOG_IDX_ITEM   IdxItem;

VOID
WriteLineNumber(
    HANDLE h,
    ULONG  Offset
    )
{
    ULONG len;

    memset(&IdxItem, 0, sizeof(IdxItem));
    IdxItem.Offset = Offset;
    WriteFile(h, &IdxItem, sizeof(IdxItem), &len, NULL);
} // end WriteLineNumber()

/*
 Catch keyboard messages
*/
DWORD
__stdcall
get_kbd(PVOID ctx)
{
    ULONG key = 0;
    ULONG tmp;
    INPUT_RECORD inp;

    if(!StdIn)
        return -1;
    while(!esc_key) {
        if(GetNumberOfConsoleInputEvents(StdIn, &key) && key) {
            key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
            if(key &&
               inp.EventType == KEY_EVENT &&
               inp.Event.KeyEvent.bKeyDown) {
                switch(inp.Event.KeyEvent.wVirtualScanCode) {
                case 0x01: /*VK_ESC*/
                    esc_key = TRUE;
                    break;
                case 0x39: /*VK_SPACE*/
                    pause_key = !pause_key;
                    print_log("Log %s\n", !pause_key ? "restarted" : "paused");
                    break;
                case 0x2e: /*VK_C*/
                    if(g_opt.output_stdout)
                        break;
                    g_opt.output_stdout = !g_opt.output_stdout;
                    break;
                case 0x31: /*VK_N*/
                    if(g_opt.output_file)
                        next_log = TRUE;
                    break;
                case 0x1f: /*VK_S*/
                    g_opt.sync_mode = !g_opt.sync_mode;
                    break;
                case 0x21: /*VK_F*/
                    flush_file = TRUE;
                    break;
                case 0x23: /*VK_H*/
                    print_log(help_text2);
                    break;
                case 0x25: /*VK_K*/
                    g_opt.skip_kernel = !g_opt.skip_kernel;
                    print_log("kernel messages %s\n", !g_opt.skip_kernel ? "on" : "off");
                    break;
                case 0x16: /*VK_Y*/
                    g_opt.skip_umode = !g_opt.skip_umode;
                    print_log("user-mode messages %s\n", !g_opt.skip_umode ? "on" : "off");
                    break;
                }
            }
        }
        Sleep(100);
    }
    return 0;
} // end get_kbd()


/*
 Catch strings from OutputDebugString()
*/
DWORD
__stdcall
get_ODS_messages(PVOID ctx)
{
    HANDLE hCompleteEvent;
    HANDLE hNewMsgEvent;
    HANDLE hdbgf;
    PDBG_OUTPUT_DEBUG_STRING_BUFFER ODSmem;
    ULONG j = 0;
    ULONG status;
    ULONG len, idx;
    PCHAR sptr;
    BOOLEAN all_copied;

    SECURITY_ATTRIBUTES SecAttr;
    SECURITY_DESCRIPTOR SecDesc;

    PDbgPrnHk_PostMessageEx_USER_IN pPutMsgBuf;

    // Connect to user-mode logger
    memset(&SecAttr, 0, sizeof(SecAttr));
    memset(&SecDesc, 0, sizeof(SecDesc));
    SecAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    SecAttr.bInheritHandle = TRUE;
    SecAttr.lpSecurityDescriptor = &SecDesc;

    if(!InitializeSecurityDescriptor(&SecDesc, SECURITY_DESCRIPTOR_REVISION)) {
        print_log("can't init SecurityDescriptor, (%#x)\n",
            GetLastError());
        return -1;
    }

    if(!SetSecurityDescriptorDacl(&SecDesc, TRUE, NULL, FALSE)) {
        print_log("can't set SD Dacl, (%#x)\n",
            GetLastError());
        return -1;
    }

    hNewMsgEvent = CreateEvent(&SecAttr, FALSE, FALSE, "DBWIN_DATA_READY");

    if(!hNewMsgEvent) {
        print_log("can't create sync event, (%#x)\n",
                GetLastError());
        return -1;
    }

    hCompleteEvent = CreateEvent(&SecAttr, FALSE, FALSE, "DBWIN_BUFFER_READY");

    if(!hCompleteEvent) {
        print_log("can't create sync event (2), (%#x)\n",
                GetLastError());
        return -1;
    }

    hdbgf = CreateFileMapping((HANDLE)(-1),
                        &SecAttr, PAGE_READWRITE,
                        0, 4096,
                        "DBWIN_BUFFER");

    if (!hdbgf) {
        print_log("can't create file mapping, (%#x)\n",
                GetLastError());
        return -1;
    }

    ODSmem = (PDBG_OUTPUT_DEBUG_STRING_BUFFER)MapViewOfFile(hdbgf, FILE_MAP_READ, 0, 0, 4096);

    if (!ODSmem) {
        print_log("can't map shared mem (%#x)\n",
                GetLastError());
        return -1;
    }

    // send Ready message
    SetEvent(hCompleteEvent);

    // inform get_messages() that ODS capturing is started
    SetEvent(hODSreadyEvent);

    if(!g_opt.without_driver) {
        pPutMsgBuf = (PDbgPrnHk_PostMessageEx_USER_IN)GlobalAlloc(GMEM_FIXED, sizeof(DbgPrnHk_PostMessageEx_USER_IN)+2048);
        if(!pPutMsgBuf) {
            print_log("can't alloc PostMessageEx buffer (%#x)\n",
                    GetLastError());
            return -1;
        }
        memset(pPutMsgBuf, 0, sizeof(DbgPrnHk_PostMessageEx_USER_IN));
        //pPutMsgBuf->TimeStamp.QuadPart = 0; //already 0
        pPutMsgBuf->ThreadId = (PVOID)(-1);
        //pPutMsgBuf->Irql = 0;             //already 0
        pPutMsgBuf->CpuNumber = 0xff;
        pPutMsgBuf->CallerMode = 1;
    }

    while(!esc_key) {
        // wait for incoming messages
        status = WaitForSingleObject(hNewMsgEvent, INFINITE);
        if(status == WAIT_OBJECT_0) {

            all_copied = FALSE;
            len = 0;

            if(g_opt.without_driver) {
retry:
                WaitForSingleObject(hODS_lock, INFINITE);

retry2:
                // copy message to buffer in unified format
                j = MsgCount2a;

                while(MsgCount2a && (MsgCount2a-1)*sizeof(DbgPrnHk_GetMessages_USER_OUT) >= BufferSize) {
                    ReleaseMutex(hODS_lock);
                    goto retry;
                };

                __try {
                    
                    idx = 0;
                    sptr = &(MsgBuffer2a[j].Msg[0]);

                    while(len < sizeof(ODSmem->Msg)) {

                        if(idx >= (sizeof(MsgBuffer2a[j].Msg)-1)) {
                            all_copied = FALSE;
                            break;
                        }
                        if(!(sptr[idx] = ODSmem->Msg[len])) {
                            all_copied = TRUE;
                            break;
                        }
                        len++;
                        idx++;
                    }
                    if(len >= sizeof(ODSmem->Msg)) {
                        all_copied = TRUE;
                    }
                    sptr[idx] = 0;

                    //strncpy(&(MsgBuffer2a[j].Msg[0]), (const char*)&(ODSmem->Msg[0]), sizeof(ODSmem->Msg));
                    //MsgBuffer2a[j].Msg[sizeof(ODSmem->Msg)-1] = 0;
                    //len = strlen(MsgBuffer2a[j].Msg);
                } __except(EXCEPTION_EXECUTE_HANDLER) {
                }
                QueryPerformanceCounter((PLARGE_INTEGER)&MsgBuffer2a[j].TimeStamp);
                MsgBuffer2a[j].Length = (USHORT)idx;
                MsgBuffer2a[j].ProcessId = (PVOID)(ODSmem->ProcessId);
                MsgBuffer2a[j].ThreadId = (PVOID)(-1);
                MsgBuffer2a[j].Irql = 0;
                MsgBuffer2a[j].CpuNumber = 0xff;
                MsgBuffer2a[j].CallerMode = 1;
                j++;
                MsgCount2a = j;

                if(!all_copied &&
                   ((MsgCount2a+1)*sizeof(DbgPrnHk_GetMessages_USER_OUT) <= BufferSize)) {
                    goto retry2;
                }

                ReleaseMutex(hODS_lock);

                if(!all_copied) {
                    goto retry;
                }
            } else {
                __try {
                    
/*
                    CHAR sptr[2048];

                    while(!all_copied) {
                        idx = 0;

                        while(len < sizeof(ODSmem->Msg)) {

                            if(idx >= (sizeof(MsgBuffer2a[j].Msg)-1)) {
                                all_copied = FALSE;
                                break;
                            }
                            if(!(sptr[idx] = ODSmem->Msg[len])) {
                                all_copied = TRUE;
                                break;
                            }
                            len++;
                            idx++;
                        }
                        if(len >= sizeof(ODSmem->Msg)) {
                            all_copied = TRUE;
                        }
                        sptr[idx] = 0;
                        DbgDump_Printn(sptr, idx);
                    }
*/
                    PUCHAR sptr = ODSmem->Msg;
                    idx = 0;
                    while((idx < sizeof(ODSmem->Msg)) && sptr[idx]) {
                        idx++;
                    }
                    pPutMsgBuf->Length = (USHORT)idx;
                    pPutMsgBuf->ProcessId = (PVOID)(ODSmem->ProcessId);
                    memcpy(pPutMsgBuf->Msg, sptr, idx);

                    DbgDump_PostMsgEx(pPutMsgBuf);

                    //strncpy(&(MsgBuffer2a[j].Msg[0]), (const char*)&(ODSmem->Msg[0]), sizeof(ODSmem->Msg));
                    //MsgBuffer2a[j].Msg[sizeof(ODSmem->Msg)-1] = 0;
                    //len = strlen(MsgBuffer2a[j].Msg);
                } __except(EXCEPTION_EXECUTE_HANDLER) {
                }
            }

            SetEvent(hCompleteEvent);
        }
    }
    return 0;
} // end get_ODS_messages()

void
dump_machine_info(
    HANDLE fh
    )
{
    CHAR Str[MAX_PATH*3];
    CHAR MHzStr[64];
    CHAR CpuId[MAX_PATH];
    CHAR VendorId[MAX_PATH];
    CHAR RegPath[MAX_PATH];
    ULONG i;
    HKEY hKey;
    ULONG MHz;
    DWORD t;
    ULONG l;
    MEMORYSTATUS          ms  ;

    for(i=0; i<128; i++) {
        sprintf(RegPath, "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\%u", i);
        if(RegOpenKey(HKEY_LOCAL_MACHINE, RegPath, &hKey) != ERROR_SUCCESS) {
            break;
        }
        // MHz
        l = sizeof(ULONG);
        if(RegQueryValueEx(hKey, "~MHz", NULL, &t, (PUCHAR)&MHz, &l) != ERROR_SUCCESS) {
            sprintf(MHzStr, "?");
        } else {
            sprintf(MHzStr, "%u", MHz);
        }
        // CPU Type
        l = MAX_PATH-1;
        if(RegQueryValueEx(hKey, "Identifier", NULL, &t, (PUCHAR)CpuId, &l) != ERROR_SUCCESS) {
            sprintf(CpuId, "Unknown type");
        } else {
            CpuId[MAX_PATH-1] = 0;
        }
        // CPU Vendor
        l = MAX_PATH-1;
        if(RegQueryValueEx(hKey, "VendorIdentifier", NULL, &t, (PUCHAR)VendorId, &l) != ERROR_SUCCESS) {
            sprintf(CpuId, "Unknown vendor");
        } else {
            VendorId[MAX_PATH-1] = 0;
        }
        l = sprintf(Str, "CPU-%u: %s, %s, %s MHz\n", i, CpuId, VendorId, MHzStr);
        TeeOutput(fh, Str, l);
        RegCloseKey(hKey);
    }
    ms.dwLength = sizeof (ms) ;
    GlobalMemoryStatus  (&ms) ;
    l = sprintf(Str, "Memory: %u Mb physical, %u Mb virtual\n", (ms.dwTotalPhys)/(1024*1024)+1, (ms.dwTotalVirtual)/(1024*1024)+1);
    TeeOutput(fh, Str, l);
}

void
dump_privileges(
    HANDLE fh
    )
{
    HANDLE              hToken;
    PTOKEN_PRIVILEGES   tp;
    DWORD               ReturnLength;
    DWORD               i, j;
    CHAR                DbgPrintBuf[MAX_PATH];
    DWORD               len;
    
    TeeOutputConst(fh, "Privilege status:\n");

    // obtain the token, first check the thread and then the process
    if (!OpenThreadToken(GetCurrentThread(), TOKEN_QUERY, TRUE, &hToken)){
        if (GetLastError() == ERROR_NO_TOKEN){
            if (!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken)) {
                TeeOutputConst(fh, "Can't open process token\n");
                return;
            }
        } else {
            TeeOutputConst(fh, "Can't open thread token\n");
            return;
        }
    }

    GetTokenInformation(hToken,TokenPrivileges, NULL,0L, &ReturnLength);
    tp = (PTOKEN_PRIVILEGES)GlobalAlloc(GMEM_FIXED,ReturnLength);
    GetTokenInformation(hToken,TokenPrivileges, tp, ReturnLength, &ReturnLength);

    for (i=0; i<tp->PrivilegeCount;i++) {
        //tp->Privileges[i].Attributes & SE_PRIVILEGE_ENABLED SE_PRIVILEGE_ENABLED_BY_DEFAULT;
        TeeOutputConst(fh, "  ");
        len = sizeof(DbgPrintBuf);
        if(LookupPrivilegeName(NULL, &tp->Privileges[i].Luid, DbgPrintBuf, &len)) {
            TeeOutput(fh, DbgPrintBuf, len);
            DbgPrintBuf[len] = 0;
        } else {
            TeeOutputConst(fh, "???");
            len = 3;
        }
        for(j=0; ((len+j+1) & ~8) < 64; j+=8) {
            TeeOutputConst(fh, "\t");
        }
        if(tp->Privileges[i].Attributes & SE_PRIVILEGE_ENABLED) {
            TeeOutputConst(fh, " Enabled");
            if(tp->Privileges[i].Attributes & SE_PRIVILEGE_ENABLED_BY_DEFAULT) {
                TeeOutputConst(fh, " (default)");
            }
        } else {
            TeeOutputConst(fh, " Disabled");
            if(!(tp->Privileges[i].Attributes & SE_PRIVILEGE_ENABLED_BY_DEFAULT)) {
                TeeOutputConst(fh, " (default");
            }
            if(DbgPrintBuf[0]) {
                if(!Privilege(DbgPrintBuf, TRUE)) {
                    TeeOutputConst(fh, ", unavailable");
                } else {
                    Privilege(DbgPrintBuf, FALSE);
                }
            }
            TeeOutputConst(fh, ")");
        }
        TeeOutputConst(fh, "\n");
    }

    TeeOutputConst(fh, "\n");

    if (!CloseHandle(hToken))
         return;

    GlobalFree(tp);
}

/*LONGLONG
translate_timestamp_to_sysperf(
    LONGLONG TimeStamp,
    UCHAR    TimeStampType,
    PDbgPrnHk_GetRdtscCalibration_USER_OUT CalibrationBuffer,
    LONGLONG* RelPerfCounter,
    double*   RdtscCalibration,
    LONGLONG* Frequency
    )*/
#define translate_timestamp_to_sysperf(TimeStamp, TimeStampType) \
{                                                                \
    switch(TimeStampType) {                                      \
    case TimeStampType_SysPerfCounter:                           \
        /*return TimeStamp;*/                                        \
        break;                                                   \
    case TimeStampType_RdtscPerfCounter:                         \
        TimeStamp = (LONGLONG)((double)(TimeStamp - CalibrationBuffer->RdtscTimeStampCalibration0) * (*RdtscCalibration)); \
        TimeStamp += CalibrationBuffer->SysTimeStampCalibration0/* - (*RelPerfCounter)*/; \
        /*return TimeStamp;*/                                        \
        break;                                                   \
    case TimeStampType_SysTime:                                  \
        TimeStamp = (TimeStamp/10000000)*(*Frequency) + (((TimeStamp%10000000) * (*Frequency)) / 10000000); \
        /*TimeStamp -= (*RelPerfCounter);*/                             \
        /*return TimeStamp;*/                                        \
        break;                                                   \
    }                                                            \
}

int
compare_timestamps(
    LONGLONG TimeStamp1,
    UCHAR    TimeStampType1,
    LONGLONG TimeStamp2,
    UCHAR    TimeStampType2,
    PDbgPrnHk_GetRdtscCalibration_USER_OUT CalibrationBuffer,
//    LONGLONG* RelPerfCounter,
    double*   RdtscCalibration,
    LONGLONG* Frequency
    )
{
    if(TimeStampType1 != TimeStampType2) {
        translate_timestamp_to_sysperf(TimeStamp1, TimeStampType1);
        translate_timestamp_to_sysperf(TimeStamp2, TimeStampType2);
    }
    if(TimeStamp1 > TimeStamp2)
        return 1;
    if(TimeStamp1 < TimeStamp2)
        return -1;
    return 0;
} // end compare_timestamps()

HANDLE
init_input_stream()
{
    ULONG k;
    HANDLE h = NULL;
    ULONG returned;
    DWORD InputMode;
    DbgPrnHk_GetVersion_USER_OUT VerBuffer;

    if(g_opt.input_driver) {
        // try to install (or re-install) and start kernel-mode DbgPrint capturing driver
        if(g_opt.restart_drv) {
            k = NtServiceStop(NT_DbgPrnHk_SVC_NAME, 5);
        }
    //    k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, PathToExec);
        k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);

        k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
        if((int)k<0) {
            k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, PathToExec);
            k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);
            k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
        }
        if(k == 1) {
            // Ok, started
        } else
        if(k == 0) {
            // installed, but not started
            k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);
            if(k < 0) {
                // disabled ?
                k = NtServiceSetStartMode(NT_DbgPrnHk_SVC_NAME, 3);
                k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);
            }
        }
        if(k < 0) {
            print_log("Warrning: Can't load or install DbgPrnHk driver\n");
            //exit(-1);
        }

        // Connect to driver
        h = CreateFile(NT_DbgPrnHk_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        if(!h || h == (HANDLE)(-1)) {
            print_log("Error: can't open driver\n");
            exit(1);
        }

        // check driver version
        if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetVersion,
                            NULL,0,
                            &VerBuffer, sizeof(VerBuffer),
                            &returned,NULL)) {
            print_log("Error: incompatible driver version\n");
            exit(1);
        }
        if(VerBuffer.Major != PROD_VER_MJ ||
           VerBuffer.Minor != PROD_VER_MN/* ||
           VerBuffer.Sub   != PROD_VER_NSUB*/) {
            print_log("Error: incompatible driver version\n");
            exit(1);
        }
    } else
    if(g_opt.input_file) {
        // Open raw message data file
        h = CreateFileW(g_opt.input_FileName, GENERIC_READ,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        if(!h || h == (HANDLE)(-1)) {
            print_log("Error: can't open input file\n");
            exit(1);
        }
    } else
    if(g_opt.input_stdin) {
        // Setup raw message data 
        if(GetConsoleMode(StdIn, &InputMode)) {
            if(!SetConsoleMode(StdIn, InputMode & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT))) {
                printf("Error: Can't set console input mode\n");
                exit(1);
            }
        }
        h = StdIn;
    } else {
        print_log("Error: no input stream found\n");
        return NULL;
    }
    return h;
} // end init_input_stream()

int
read_input_stream(
    HANDLE h,
    PVOID MsgBuffer,
    ULONG BufferSize
    )
{
    PDbgPrnHk_GetMessages_USER_OUT tmp_MsgBuffer2;
    ULONG returned;

    if(g_opt.input_driver) {
/*
        // Read calibration info
        ReCalibrationBuffer.Recalibrate = TRUE;
        ReCalibrationBuffer.ChangeType = FALSE;
        if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetRdtscCalibration,
                            &ReCalibrationBuffer, sizeof(ReCalibrationBuffer),
                            &CalibrationBuffer, sizeof(CalibrationBuffer),
                            &returned,NULL)) {
            returned = GetLastError();
            returned = 0;
            RdtscCalibration = 1.0;
        } else {
            RdtscCalibration = (double)(CalibrationBuffer.SysTimeStampCalibration - CalibrationBuffer.SysTimeStampCalibration0) /
                               (double)(CalibrationBuffer.RdtscTimeStampCalibration - CalibrationBuffer.RdtscTimeStampCalibration0);
        }
*/
        // Get kernel messages
        if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetMessages,
                            MsgBuffer, min(BufferSize, 0x10000),
                            MsgBuffer, min(BufferSize, 0x10000),
                            &returned,NULL)) {
            returned = GetLastError();
            returned = 0;
        }
        if(g_opt.without_driver) {
            // Get user-mode messages
            WaitForSingleObject(hODS_lock, INFINITE);

            // swap read/write buffers
            tmp_MsgBuffer2 = MsgBuffer2;

            MsgCount2   = MsgCount2a;
            MsgBuffer2  = MsgBuffer2a;

            MsgCount2a  = 0;
            MsgBuffer2a = tmp_MsgBuffer2;

            ReleaseMutex(hODS_lock);
        }

    } else
    if(g_opt.input_file) {
        // Read messages
        if(!ReadFile(h,
                            MsgBuffer, BufferSize,
                            &returned,NULL) ||
           !returned) {
            returned = GetLastError();
            returned = 0;
            esc_key = TRUE;
        }
    } else
    if(g_opt.input_stdin) {
        // Read messages from pipe
        returned = 0;
        if(!ReadFile(h,
//                                    MsgBuffer, min(BufferSize, returned),
                            MsgBuffer, BufferSize,
                            &returned,NULL)) {
            returned = GetLastError();
            returned = 0;
            esc_key = TRUE;
        }
    }
    return returned;
} // end read_input_stream()

void
close_input_stream(
    HANDLE h
    )
{
    if(!g_opt.input_stdin) {
        CloseHandle(h);
    }
} // end close_input_stream()

/*
 merge user- and kernel-mode logs
*/
void
get_messages()
{
    HANDLE h;
    HANDLE fh;
    HANDLE ih;

    ULONG len;
    ULONG i, j, j2;
//    ULONG k;
//    ULONG r;
    WCHAR a;
    PWCHAR NumPos;
    ULONG returned;
    BOOLEAN new_line;
    BOOLEAN CallerMode;
    CHAR DbgPrintBuf[4096];
    ULONG SleepCount;
    CHAR tab3[] = "\t\t\t";
    ULONG CurLogSize;
    SYSTEMTIME SysTime;

    OSVERSIONINFO OsVer;
    PCHAR PlatformStr;

    DBGPRN_FORMAT_MESSAGE_CTX msgst;
    DBGPRN_TIMESTAMP_CTX      timest;

    HANDLE hODSthread = NULL;
    ULONG thId;
    HANDLE handles[3];
    
    TIME_ZONE_INFORMATION TzInfo;
    LONG Bias;

    BOOLEAN last2 = FALSE;
    PDbgPrnHk_GetMessages_USER_OUT CurMsg = NULL;

    //DbgPrnHk_GetRdtscCalibration_USER_OUT CalibrationBuffer;
    //double RdtscCalibration;

    // gain privileges to install/load driver/service
    // it is not necessary here if we are not running as service
    if(sys_account) {
        if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
            print_log("Warning: Insufficient privileges\n");
        }
    }

    h = init_input_stream();

    if(g_opt.output_file) {
        // Init log-file name
        wcsncpy(LogFileN, g_opt.LogFile, MAX_PATH-10);
        LogFileN[MAX_PATH-10] = 0;
        len = wcslen(LogFileN);

        NumPos = &LogFileN[len];
        LogFileSuf[0] = 0;
        for(i=len; i>0; i--) {
            a = LogFileN[i-1];
            if(a == '\\') {
                break;
            }
            if(a == '.') {
                NumPos = &LogFileN[i-1];
                wcscpy(LogFileSuf, NumPos);
                break;
            }
        }
    }

    if(!g_opt.output_none) {
        // alloc buffers
        MsgBuffer = (PDbgPrnHk_GetMessages_USER_OUT)
            GlobalAlloc(GMEM_FIXED, BufferSize);
        if(!MsgBuffer) {
            print_log("Error: can't allocate message buffer\n");
            exit(4);
        }
        MsgBuffer2 = (PDbgPrnHk_GetMessages_USER_OUT)
            GlobalAlloc(GMEM_FIXED, BufferSize);
        if(!MsgBuffer2) {
            print_log("Error: can't allocate message buffer (2)\n");
            exit(4);
        }
        MsgBuffer2a = (PDbgPrnHk_GetMessages_USER_OUT)
            GlobalAlloc(GMEM_FIXED, BufferSize);
        if(!MsgBuffer2a) {
            print_log("Error: can't allocate message buffer (2)\n");
            exit(4);
        }
    }
    i = g_opt.StartLogNum;

    if(!g_opt.input_stdin) {
        print_log(help_text2);
    }

    if(sys_account) {
        // change service status to RUNNING
        ReportStatusToSCMgr(
            SERVICE_RUNNING,      // service state
            NO_ERROR,             // exit code
            0);                   // wait hint
    }

    if(g_opt.input_driver) {
        hODSreadyEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
        if(!hODSreadyEvent) {
            print_log("Error: can't create event\n");
            exit(4);
        }
        // start user-mode OutputDebugString capturing thread
        hODSthread = CreateThread(NULL, 0, get_ODS_messages, NULL, 0, &thId);
        if(!hODSthread) {
            print_log("Error: can't create thread\n");
            exit(4);
        }
        handles[0] = hODSreadyEvent;
        handles[1] = hODSthread;
        if(WaitForMultipleObjects(2, handles, FALSE, 3*1000) != WAIT_OBJECT_0) {
            print_log("Error: can't init OutputDebugString() capturing thread\n");
            exit(4);
        }
    }
    // produce logs until break come
    while(!esc_key) {

        if(g_opt.output_none) {
            Sleep(100);
            continue;
        }
        if(g_opt.output_file) {
            // find unused name for log-file
            if(i>0) {
                swprintf(NumPos, L"%d%s", i, LogFileSuf);
            }
            fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
            if(fh && fh != (HANDLE)(-1)) {
                CloseHandle(fh);
                fh = NULL;
                i++;
                if(!i) {
                    print_log("Error: find unused filename\n");
                    exit(2);
                }
                continue;
            }
            if((i >= g_opt.MaxLogs) && (i-g_opt.MaxLogs >= g_opt.StartLogNum)) {
                if(i-g_opt.MaxLogs > 0) {
                    swprintf(NumPos, L"%d%s", i-g_opt.MaxLogs, LogFileSuf);
                } else {
                    swprintf(NumPos, L"%s", LogFileSuf);
                }
                DeleteFileW(LogFileN);
                if(i > 0) {
                    swprintf(NumPos, L"%d%s", i, LogFileSuf);
                } else {
                    swprintf(NumPos, L"%s", LogFileSuf);
                }
            }
            // create log-file
            fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           CREATE_NEW,
                           FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
            if(g_opt.use_index) {
                LineNumber = 0;
                swprintf(IdxFileN, L"%s.idx", LogFileN);
                ih = CreateFileW(IdxFileN, GENERIC_READ | GENERIC_WRITE,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               CREATE_NEW,
                               FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
                memset(&IdxHeader, 0, sizeof(IdxHeader));

                IdxHeader.log_pid       = g_opt.log_pid      ;
                IdxHeader.log_tid       = g_opt.log_tid      ;
                IdxHeader.log_mode      = g_opt.log_mode     ;
                IdxHeader.log_irql      = g_opt.log_irql     ;
                IdxHeader.log_time_ex   = g_opt.log_time_ext ;
                IdxHeader.log_time_rel  = g_opt.log_time_perf ;
                IdxHeader.log_time_date = g_opt.log_time_date;
                IdxHeader.log_time_time = g_opt.log_time_time;
                             
                IdxHeader.skip_kernel   = g_opt.skip_kernel  ;
                IdxHeader.skip_umode    = g_opt.skip_umode   ;

                IdxHeader.log_time_nano = g_opt.log_time_nano;

                IdxHeader.Major  = IDX_VER_MJ;
                IdxHeader.Minor  = IDX_VER_MN;
                IdxHeader.Sub    = IDX_VER_NSUB;

                IdxHeader.Length = sizeof(IdxHeader);
                IdxHeader.IdxItemSize = sizeof(IdxItem);
            }
        } else {
            fh = StdOut;
        }
        if(!fh || fh == (HANDLE)(-1)) {
            print_log("Error: can't create log-file\n");
            fh = NULL;
            exit(3);
        }
        if(g_opt.use_index) {
            if(!ih || ih == (HANDLE)(-1)) {
                print_log("Error: can't create index-file\n");
                ih = NULL;
                exit(3);
            }
            WriteFile(ih, &IdxHeader, sizeof(IdxHeader), &len, NULL);
        }
        if(g_opt.prealloc_mode) {
            SetFilePointer(fh, g_opt.MaxLogSize*1024*1024, NULL, FILE_BEGIN);
            SetEndOfFile(fh);
            SetFilePointer(fh, 0, NULL, FILE_BEGIN);
        }

        dbgprint_format_msg_init(&g_opt, &msgst);

        // write log header
        if(g_opt.output_file) {
            print_log("Writing log to %ws\n", LogFileN);
        }
        CurLogSize = 0;
        if(g_opt.input_driver) {
            len = sprintf(DbgPrintBuf, "Log generated by DbgPrintLog v%d.%d%s\n", PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
            TeeOutput(fh, DbgPrintBuf, len);

            // init timestamp calibration
            dbgprint_timestamp_init(&g_opt, h, &timest);

            FileTimeToSystemTime((PFILETIME)&timest.FtmSysTime, &SysTime);
            len = sprintf(DbgPrintBuf, "Timestamp: %d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d (%I64d.%4.4d ticks), %s\n",
                SysTime.wYear,
                SysTime.wMonth,
                SysTime.wDay,
                SysTime.wHour,
                SysTime.wMinute,
                SysTime.wSecond,
                timest.RelPerfCounter/10000,
                (ULONG)(timest.RelPerfCounter%10000),
                g_opt.log_time_utc ? "UTC" : "Local"
                );
            TeeOutput(fh, DbgPrintBuf, len);

            GetTimeZoneInformation(&TzInfo);
            Bias = TzInfo.Bias + TzInfo.DaylightBias;
            len = sprintf(DbgPrintBuf, "Time Zone: %s%2.2d%2.2d, %S\n",
                Bias >= 0 ? "-" : "+",
                abs(Bias) / 60,
                abs(Bias) % 60,
                TzInfo.StandardName
                );
            TeeOutput(fh, DbgPrintBuf, len);

            OsVer.dwOSVersionInfoSize = sizeof(OsVer);
            GetVersionEx(&OsVer);
            switch(OsVer.dwPlatformId) {
            case VER_PLATFORM_WIN32s:
                PlatformStr = "Win";
                break;
            case VER_PLATFORM_WIN32_WINDOWS:
                PlatformStr = "Win 9x";
                break;
            case VER_PLATFORM_WIN32_NT:
                PlatformStr = "NT";
                break;
            default:
                PlatformStr = "Unknown";
            }
            len = sprintf(DbgPrintBuf, "OS Version: %d:%s %d.%d (%d)\n  %s\n",
                OsVer.dwPlatformId,
                PlatformStr,
                OsVer.dwMajorVersion,
                OsVer.dwMinorVersion,
                OsVer.dwBuildNumber,
                OsVer.szCSDVersion);
            TeeOutput(fh, DbgPrintBuf, len);

            TeeOutputConst(fh, "Hostname: ");
            len = 1023;
            GetComputerName(DbgPrintBuf, &len);
            DbgPrintBuf[1023] = 0;
            TeeOutput(fh, DbgPrintBuf, len);

            TeeOutputConst(fh, "\nUsername: ");
            len = 1023;
            GetUserName(DbgPrintBuf, &len);
            DbgPrintBuf[1023] = 0;
            TeeOutput(fh, DbgPrintBuf, len);

            TeeOutputConst(fh, "\n");
            dump_machine_info(fh);
            TeeOutputConst(fh, "\n");
            dump_privileges(fh);
            TeeOutputConst(fh, "\n");
        } else {
            len = sprintf(DbgPrintBuf, "Log translated by DbgPrintLog v%d.%d%s\n", PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
            TeeOutput(fh, DbgPrintBuf, len);
        }
        TeeOutputConst(fh, "----- End of Log header -----\n");

        // write to single log-file
        next_log = FALSE;
        while(!esc_key && !next_log) {

            dbgprint_timestamp_resync(&g_opt, h, &timest);
            
            returned = read_input_stream(h, MsgBuffer, BufferSize);
/*
            if(g_opt.skip_kernel) {
                returned = 0;
            }
*/
            if(g_opt.skip_umode) {
                MsgCount2 = 0;
            }

            // drop messages if paused
            if(pause_key) {
                returned = 0;
                MsgCount2 = 0;
            }
            if(!returned && !MsgCount2) {
                SleepCount++;
                if(SleepCount > g_opt.FlushTimeout) {
                    SleepCount = 0;
                    new_line = TRUE;
                    FlushFileBuffered(fh);
                }
                Sleep(100);
                continue;
            }
            SleepCount = 0;
            j = 0;
            j2 = 0;
            // join user- and kernel-mode logs
            while(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) || MsgCount2) {
                // assure time ordering
                if(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) && MsgCount2) {
/*                    if(MsgBuffer [j] .TimeStamp.QuadPart >
                       MsgBuffer2[j2].TimeStamp.QuadPart) {*/
                      if(compare_timestamps(
                             MsgBuffer [j] .TimeStamp.QuadPart,
                             MsgBuffer [j] .TimeStampType,
                             MsgBuffer2[j2].TimeStamp.QuadPart,
                             MsgBuffer2[j2].TimeStampType,
                             &timest.CalibrationBuffer,
                             /*&RelPerfCounter,*/
                             &timest.RdtscCalibration,
                             (PLONGLONG)&timest.Frequency
                                           )
                          > 0) {
                        CurMsg = &MsgBuffer2[j2];
                        if(!last2) {
                            msgst.new_line = TRUE;
                            last2 = TRUE;
                        }
                    } else {
                        CurMsg = &MsgBuffer[j];
                        if(last2) {
                            msgst.new_line = TRUE;
                            last2 = FALSE;
                        }
                    }
                } else
                if(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT)) {
                    CurMsg = &MsgBuffer[j];
                    if(last2) {
                        new_line = TRUE;
                        last2 = FALSE;
                    }
                } else {
                    CurMsg = &MsgBuffer2[j2];
                    if(!last2) {
                        msgst.new_line = TRUE;
                        last2 = TRUE;
                    }
                }
                // filter message by origin mode
                CallerMode = CurMsg->CallerMode;
                if(g_opt.skip_kernel && !CallerMode) {
                    CurMsg->Length = 0;
                }
                if(g_opt.skip_umode && CallerMode) {
                    CurMsg->Length = 0;
                }
                // write message to file (buffered)
                len = dbgprint_format_msg(&g_opt, CurMsg, DbgPrintBuf, &timest, &msgst);
                if(len) {
                    if(msgst.new_line) {
                        LineNumber++;
                        WriteLineNumber(ih, CurLogSize);
                    }
                    TeeOutput(fh, DbgPrintBuf, len);
                    CurLogSize+=len;
                }

                // go to next message
                if(last2) {
                    j2++;
                    MsgCount2--;
                } else {
                    j++;
                    returned -= sizeof(DbgPrnHk_GetMessages_USER_OUT);
                }
            } // end while(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) || MsgCount2)
            if(j+j2 < 256) {
                Sleep(20);
            }
            // check user controls
            if(flush_file || g_opt.sync_mode) {
                if(g_opt.output_file && !g_opt.sync_mode) {
                    print_log("flush log\n");
                }
                FlushFileBuffered(fh);
                flush_file = FALSE;
            }
            if(next_log) {
                print_log("next log\n");
            }
            if((CurLogSize+0x10000*2) > g_opt.MaxLogSize*1024*1024 && g_opt.output_file) {
                next_log = TRUE;
                print_log("max log size limit reached -> next log\n");
            }
        } // end of Get Message loop
        FlushFileBuffered(fh);
        CloseHandle(fh);
        fh = NULL;
        if(g_opt.use_index) {
            CloseHandle(ih);
            ih = NULL;
        }
    }

    // cleanup
    if(fh) {
        CloseHandle(fh);
    }
    close_input_stream(h);
} // end get_messages()

static const char help_text[] = 
    "Usage:\n"
    "\tDbgPrintLog.exe\t[<switches>] [<log-file name>]\n"
    "Switches:\n"
    "    Log-file format:\n"
    "\t-m       write initiator mode (K - kernel, U - user) to log\n"
    "\t-p       write ProcessId to log\n"
    "\t-t       write ThreadId to log (for kernel-mode only)\n"
    "\t-i       write IRQL to log (for kernel-mode only)\n"
    "\t-cpu     write CPU num to log (for kernel-mode only)\n"
    "\t--full   same as -m -p -t -i\n"
    "\t-T FMT   specify absolute time format FMT. FMT string can contain\n"
    "\t         the following switches: D - date, T - time, N - high precision time\n"
    "\t         R - relative time (tick count), U - UTC time\n"
//    "\t+fp PID  log only messages with ProcessId==PID\n"
//    "\t-fp PID  do not log messages with ProcessId==PID\n"
    "\t-fm M    do not log messages from <M> mode\n"
    "\t         <M> can be K - kernel or U - user\n"
    "    Log-file control:\n"
    "\t-s NUM   set max log-file size to NUM Mbytes (0 < NUM < 2048)\n"
    "\t         128Mb is used by default\n"
    "\t-S NUM   does like -s, but sets initial file size to specified value.\n"
    "\t         Is intended mainly for use by GUI tool.\n"
    "\t-x NUM   create index file for each log file.\n"
    "\t         Is intended mainly for use by GUI tool.\n"
    "\t-l NUM   keep NUM latest log files\n"
    "\t-n NUM   set start log number to NUM\n"
    "\t-ft NUM  flush messages buffer after NUM seconds of inactivity\n"
    "\t-sm      synchronous mode. all messages are flushed immediately\n"
    "\t-cf      copy log to both STDOUT and file\n"
    "\t-wd DIR  specify working directory\n"
    "    Output data stream control:\n"
    "\t-o\n"
    "\t   or\n"
    "\t--stdout write log to STDOUT instead of file\n"
    "\t--no_out do not write logs anyware\n"
    "\t         just forward OutputDebugString to driver\n"
    "    Input data stream control:\n"
    "\t--stdin  read message stream from STDIN\n"
    "\t--in_file FILENAME\n"
    "\t         read message stream from FILENAME\n"
    "\t--in_drv read message stream from the driver (default)\n"
    "\t--no_drv do not use driver for message routing\n"
    "    Run mode:\n"
    "\t-svc MOD install or uninstall as service. MOD specify startup mode:\n"
    "\t         A - automatic, M - manual, U - uninstall. 'A' is used by default\n"
    "\t         if MOD is omited.\n"
    "\t-drv MOD Specify driver startup mode:\n"
    "\t         1 - very first (boot), B - boot, S - system,\n"
    "\t         A - automatic, M - manual, U - uninstall\n"
    "\t         'M' is used by default if MOD is omited.\n"
    "\t--drvopt OPTION_NAME VALUE\n"
    "\t         Specify driver startup option. Valid OPTION_NAMEs are\n"
    "\t             CheckIrql, BufferSize, DoNotPassMessagesDown,\n"
    "\t             StopOnBufferOverflow, TimeStampType\n"
    "\t         Read documentation for each option description\n"
    "\t-rd      restart and reinstall driver before start\n"
//    "\t-chcfg   Make config changes only and exit. Do not start log capturing\n"
    "\t-h       display this help message\n"
//    "\t-h OPT   display help for driver option OPT\n"
    ;

void
usage(void)
{
    HANDLE stdOuth = GetStdHandle(STD_OUTPUT_HANDLE);
    ULONG WrittenBytes;

    if(stdOuth && stdOuth != ((HANDLE)(-1))) {
        WriteFile(stdOuth, help_text,  strlen(help_text),  &WrittenBytes, NULL);
        WriteFile(stdOuth, help_text2, strlen(help_text2), &WrittenBytes, NULL);
    } else {
        MessageBox(NULL, help_text, "Usage", MB_OK);
    }
} // end usage()

BOOL
ReportStatusToSCMgr(
    DWORD dwCurrentState,
    DWORD dwWin32ExitCode,
    DWORD dwWaitHint
    )
{
    static DWORD dwCheckPoint = 1;
    BOOL fResult = TRUE;

    if (dwCurrentState == SERVICE_START_PENDING) {
        ssStatus.dwControlsAccepted = 0;
    } else {
        ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_PAUSE_CONTINUE;
    }

    ssStatus.dwCurrentState = dwCurrentState;
    ssStatus.dwWin32ExitCode = dwWin32ExitCode;
    ssStatus.dwWaitHint = dwWaitHint;

    if ( ( dwCurrentState == SERVICE_RUNNING ) ||
         ( dwCurrentState == SERVICE_STOPPED ) ) {
        ssStatus.dwCheckPoint = 0;
    } else {
        ssStatus.dwCheckPoint = dwCheckPoint++;
    }

    // Report the status of the service to the service control manager.
    if (!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus))) {
        AddToMessageLog(TEXT("SetServiceStatus"));
    }
    return fResult;
} // end ReportStatusToSCMgr()

VOID
AddToMessageLog(
    LPTSTR lpszMsg
    )
{
    CHAR    szMsg[256];
    HANDLE  hEventSource;
    LPSTR   lpszStrings[2];
    ULONG   dwErr;

    dwErr = GetLastError();

    // Use event logging to log the error.
    hEventSource = RegisterEventSource(NULL, NT_DbgPrnHk_Client_SVC_NAME);

    sprintf(szMsg, "%s error: %d", NT_DbgPrnHk_Client_SVC_NAME, dwErr);
    lpszStrings[0] = szMsg;
    lpszStrings[1] = lpszMsg;

    if (!hEventSource)
        return;

    ReportEvent(hEventSource, // handle of event source
        EVENTLOG_ERROR_TYPE,  // event type
        0,                    // event category
        0,                    // event ID
        NULL,                 // current user's SID
        2,                    // strings in lpszStrings
        0,                    // no bytes of raw data
        (const char **)lpszStrings,          // array of error strings
        NULL);                // no raw data

    (VOID)DeregisterEventSource(hEventSource);
} // end AddToMessageLog()

void
WINAPI
service_ctrl(
    DWORD dwCtrlCode)
{
    ULONG sst;
    // Handle the requested control code.
    switch(dwCtrlCode)
    {
    // Stop the service.

    // SERVICE_STOP_PENDING should be reported before
    // setting the Stop Event - hServerStopEvent - in
    // ServiceStop().  This avoids a race condition
    // which may result in a 1053 - The Service did not respond...
    // error.
    case SERVICE_CONTROL_STOP:
        ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 3000);
        esc_key = TRUE;
        break;

    case SERVICE_CONTROL_PAUSE:
        ReportStatusToSCMgr(SERVICE_PAUSED, NO_ERROR, 0);
        pause_key = TRUE;
        break;

    case SERVICE_CONTROL_CONTINUE:
        ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 0);
        pause_key = FALSE;
        break;

    default:
    case SERVICE_CONTROL_INTERROGATE:
        if(esc_key) {
            sst = SERVICE_STOP_PENDING;
        } else
        if(pause_key) {
            sst = SERVICE_PAUSED;
        } else {
            sst = SERVICE_RUNNING;
        }
        ReportStatusToSCMgr(sst, NO_ERROR, 0);
        break;
    }
} // service_ctrl()

void
WINAPI
service_main(
    DWORD dwArgc,
    LPTSTR *lpszArgv)
{
    // register our service control handler:
    sshStatusHandle = RegisterServiceCtrlHandler( NT_DbgPrnHk_Client_SVC_NAME, service_ctrl);
    print_log("***** DbgPrintLog service start********");
    if (!sshStatusHandle)
        goto cleanup;

    // SERVICE_STATUS members that don't change in example
    ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwServiceSpecificExitCode = 0;


    // report the status to the service control manager.
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    // get messages from user- and kernel-mode and store them to log-file
    get_messages();

cleanup:

    // try to report the stopped status to the service control manager.
    if (sshStatusHandle)
        (VOID)ReportStatusToSCMgr(
                            SERVICE_STOPPED,
                            NO_ERROR,
                            0);

    return;
} // end service_main()

BOOLEAN
deinstall_very_first()
{
    int k;
    HKEY hKey;
    PULONG v;
    ULONG t;
    int i;
    BOOLEAN updated = FALSE;

    // save load order in registry
    if(RegOpenKeyW(HKEY_LOCAL_MACHINE, L"SYSTEM\\CurrentControlSet\\Control\\GroupOrderList", &hKey) == ERROR_SUCCESS) {
        k = 0;
        RegQueryValueExW(hKey, L"System Bus Extender", NULL, &t, NULL, (PULONG)&k);
        v = (PULONG)GlobalAlloc(GMEM_FIXED, k);
        RegQueryValueExW(hKey, L"System Bus Extender", NULL, &t, (PUCHAR)v, (PULONG)&k);
        i = 1;
        while(i*(int)sizeof(ULONG) < k) {
            if(v[i] == NT_DBGPRNHK_DRV_TAG) {
                v[0]--;
                memmove(&(v[i]), &(v[i+1]), k-(i+1)*sizeof(ULONG));
                k -= 4;
                updated = TRUE;
            } else {
                i++;
            }
        }
        if(updated) {
            if(RegSetValueExW(hKey, L"System Bus Extender", NULL, REG_BINARY, (PUCHAR)v, k-sizeof(ULONG)) != ERROR_SUCCESS) {
                print_log("Error: Can't store driver load order in registry\n");
            }
        }
        GlobalFree(v);
        RegCloseKey(hKey);
    } else {
        print_log("Error: Can't store driver load order in registry\n");
        return FALSE;
    }
    return TRUE;
} // end deinstall_very_first()

BOOLEAN
remove_driver()
{
    int k;
    HANDLE h;
    ULONG returned;

    deinstall_very_first();

    k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
    if(k<0) {
        NtServiceSetStartMode(NT_DbgPrnHk_SVC_NAME, 4);
        return TRUE;
    }
    if(k == 1) {
        // Started, connect to driver
        h = CreateFile(NT_DbgPrnHk_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        if(!h || h == (HANDLE)(-1)) {
            print_log("Error: can't open driver\n");
        } else {
            // check if driver can be unloaded
            if(!DeviceIoControl(h,IOCTL_DbgPrnHk_CanUnload,
                                NULL,0,
                                NULL,0,
                                &returned,NULL)) {
                print_log("Warning: DbgPrnHk driver cannot be unloaded, will be disabled\n");
                NtServiceSetStartMode(NT_DbgPrnHk_SVC_NAME, 4);
                CloseHandle(h);
                return FALSE;
            } else {
                k = NtServiceStop(NT_DbgPrnHk_SVC_NAME, 1);
            }
        }
    } else
    if(k == 0) {
        NtServiceSetStartMode(NT_DbgPrnHk_SVC_NAME, 4);
    }

    k = NtServiceRemove(NT_DbgPrnHk_SVC_NAME);
    if(k == 1) {
        print_log("DbgPrnHk driver is uninstalled\n");
        return TRUE;
    }
    return FALSE;
} // end remove_driver()

BOOLEAN
AmIStarted()
{
    HANDLE h;
    ULONG ErrCode;

    h = CreateEvent(NULL, FALSE, FALSE, "DbgPrintLogEvent1");
    if(!h)
        return FALSE;
    ErrCode = GetLastError();
    if(ErrCode == ERROR_ALREADY_EXISTS)
        return TRUE;
    return FALSE;
} // end AmIStarted()

void main () {
//    int argc;
//    WCHAR** argv;
    WCHAR*  CmdLine = NULL;
    WCHAR*  CmdLine2;

    LONG i=0;
//    LONG j;
    LONG k;
    ULONG t;
//    WCHAR a;
    ULONG thId;
    HKEY hKey;
    PCHAR sPtr;
    PWCHAR tmpWDir = NULL;

    print_log("DbgPrintLog v%d.%d%s (c) by Alter\nHome site http://www.alter.org.ua\n",
        PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
    fflush(stdout);

    if(GetVersion() & 0x80000000) {
        print_log("For Windows NT family only.\n");
        exit(-2);
    }

    // init with default settings
    #define DBGPRINT_OPT_RAW_INIT
    #include "dbgprint_opt_list.h"

//    __asm int 3;

    // check if we are running as service
    sys_account = IsSystemAccount();

    // get current directory
    k = GetCurrentDirectoryW(0, NULL);
    if(k) {
        g_opt.WDir = (PWCHAR)GlobalAlloc(GMEM_FIXED, (k+1)*sizeof(WCHAR));
        g_opt.WDir[0] = 0;
        k = GetCurrentDirectoryW(k, g_opt.WDir);
        g_opt.WDir[k] = 0;
    }

    if(sys_account) {
        // if we are running as service get command line and working directory from registry

/*        k = MAX_PATH*3;
        CmdLine = (PWCHAR)GlobalAlloc(GMEM_FIXED, k);
        if(!CmdLine) {
            exit(-3);
        }*/
        if(RegOpenKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\" NT_DbgPrnHk_Client_SVC_NAME, &hKey) != ERROR_SUCCESS) {
            print_log("Can't open %S registry key\n", "System\\CurrentControlSet\\Services\\" NT_DbgPrnHk_Client_SVC_NAME);
            exit(-4);
        }
        k = 0;
        if(RegQueryValueExW(hKey, NT_DbgPrnHk_Reg_CmdLineW, NULL, &t, NULL, (PULONG)&k) == ERROR_SUCCESS) {
            CmdLine = (PWCHAR)GlobalAlloc(GMEM_FIXED, k+sizeof(WCHAR));
        }
        if(!k || ! CmdLine ||
           RegQueryValueExW(hKey, NT_DbgPrnHk_Reg_CmdLineW, NULL, &t, (PUCHAR)CmdLine, (PULONG)&k) != ERROR_SUCCESS) {
            if(CmdLine)
                GlobalFree(CmdLine);
            CmdLine = GetCommandLineW();
        }
        k = 0;
        if(RegQueryValueExW(hKey, NT_DbgPrnHk_Reg_WDirW, NULL, &t, NULL, (PULONG)&k) == ERROR_SUCCESS) {
            tmpWDir = (PWCHAR)GlobalAlloc(GMEM_FIXED, k+sizeof(WCHAR));
        }
        if(k && tmpWDir && RegQueryValueExW(hKey, NT_DbgPrnHk_Reg_WDirW, NULL, &t, (PUCHAR)tmpWDir, (PULONG)&k) == ERROR_SUCCESS) {
            g_opt.WDir = tmpWDir;
        } else {
            if(tmpWDir)
                GlobalFree(tmpWDir);
        }
        RegCloseKey(hKey);
    } else {
        CmdLine = GetCommandLineW();
        if(!CmdLine) {
            exit(-1);
        }
    }

    GetModuleFileName(NULL, PathToExec, sizeof(PathToExec));
    sPtr = strrchr(PathToExec, '\\');
    if(sPtr) {
        (*sPtr) = 0;
    }

    k = DbgPrint_CmdLine_to_OptStruct(CmdLine, FALSE, (PDBGDUMP_OPT)&g_opt);
    if(k != 1) {
        if(!k)
            usage();
        exit(k);
    }


    StdIn  = GetStdHandle(STD_INPUT_HANDLE);
    StdOut = GetStdHandle(STD_OUTPUT_HANDLE);

    // gain privileges to install/load driver/service
    if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
        print_log("Warning: Insufficient privileges\n");
    }

//    if(sys_account) {
//        continue;
//    }

/*
    // do nothing if this was service/driver deinstall
    if(deinstall) {
        exit(0);
    }
*/
    // set current directory
    SetCurrentDirectoryW(g_opt.WDir);

    if(!sys_account && g_opt.input_driver) {
        // ensure that driver is installed

        if(g_opt.install_svc && g_opt.svc_mode != SERVICE_DISABLED) {
            g_opt.install_drv = TRUE;
        }

        if(g_opt.install_drv) {
            // install/deinstall or update startup mode
            if(g_opt.drv_mode == SERVICE_DISABLED) {
                remove_driver();
            } else {
                k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
                if(k<0) {
                    remove_driver();
                    k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, PathToExec, TRUE, g_opt.drv_mode);
                    if(k < 0) {
                        print_log("Error: Can't install DbgPrnHk driver\n");
                        exit(-1);
                    }
                } else {
                    // update .SYS
                    NtServiceInstall(NT_DbgPrnHk_SVC_NAME, PathToExec, TRUE, g_opt.drv_mode);
                    // update start mode
                    NtServiceSetStartMode(NT_DbgPrnHk_SVC_NAME, g_opt.drv_mode);
                }
                // save load options in registry
                if(RegOpenKeyW(HKEY_LOCAL_MACHINE, NT_DBGPRNHK_REG_PATH, &hKey) != ERROR_SUCCESS) {
                    print_log("Error: Can't store driver parameters in registry\n");
                    exit(-4);
                }
                k = sizeof(ULONG);
                i = g_opt.drv_mode; // load mode
                if(RegSetValueExW(hKey, L"Start", NULL, REG_DWORD, (PUCHAR)&i, k) != ERROR_SUCCESS) {
                    print_log("Error: Can't store driver parameters in registry\n");
                    exit(-4);
                }
                if(g_opt.drv_very_first) {
                    PULONG v = NULL;

                    // make clean
                    deinstall_very_first();

                    // make install
                    // ;)
                    k = sizeof(ULONG);
                    i = NT_DBGPRNHK_DRV_TAG; // our Tag
                    if(RegSetValueExW(hKey, NT_DBGPRNHK_REG_TAG_VAL, NULL, REG_DWORD, (PUCHAR)&i, k) != ERROR_SUCCESS) {
                        print_log("Error: Can't store driver parameters in registry\n");
                        exit(-4);
                    }
                    k = sizeof(L"System Bus Extender");
                    if(RegSetValueExW(hKey, NT_DBGPRNHK_REG_GROUP_VAL, NULL, REG_SZ, (PUCHAR) L"System Bus Extender", k) != ERROR_SUCCESS) {
                        print_log("Error: Can't store driver parameters in registry\n");
                        exit(-4);
                    }
        /*
                    // do not set TSTAMP_TYPE directly to override explicitly set value (if any)
                    for(j=0; DbgPrintHook_Options[j].OptionName; j++) {
                        if(!wcsicmp(DbgPrintHook_Options[j].OptionName, NT_DBGPRNHK_REG_TSTAMP_TYPE_VAL)) {
                            DbgPrintHook_Options[j].Value.OptDword = TimeStampType_RdtscPerfCounter;
                            DbgPrintHook_Options[j].UpdateFlag = TRUE;
                            drv_opt = TRUE;
                            break;
                        }
                    }
        */
                    g_opt.TimeStampType = TimeStampType_RdtscPerfCounter;
                    g_opt.TimeStampType_upd = TRUE;
                    g_opt.drv_opt = TRUE;

                    RegCloseKey(hKey);
                    // save load order in registry
                    if(RegOpenKeyW(HKEY_LOCAL_MACHINE, L"SYSTEM\\CurrentControlSet\\Control\\GroupOrderList", &hKey) != ERROR_SUCCESS) {
                        print_log("Error: Can't store driver load order in registry\n");
                        exit(-4);
                    }
                    k = 0;
                    RegQueryValueExW(hKey, L"System Bus Extender", NULL, &t, NULL, (PULONG)&k);
                    v = (PULONG)GlobalAlloc(GMEM_FIXED, k+sizeof(ULONG));
                    RegQueryValueExW(hKey, L"System Bus Extender", NULL, &t, (PUCHAR)v, (PULONG)&k);
                    if(v[1] != NT_DBGPRNHK_DRV_TAG) {
                        v[0]++;
                        memmove(&(v[2]), &(v[1]), k-sizeof(ULONG));
                        v[1] = NT_DBGPRNHK_DRV_TAG;
                        if(RegSetValueExW(hKey, L"System Bus Extender", NULL, REG_BINARY, (PUCHAR)v, k+sizeof(ULONG)) != ERROR_SUCCESS) {
                            print_log("Error: Can't store driver load order in registry\n");
                            exit(-4);
                        }
                    }
                    GlobalFree(v);
                } else {
                    RegDeleteValueW(hKey, NT_DBGPRNHK_REG_TAG_VAL);
                    RegDeleteValueW(hKey, NT_DBGPRNHK_REG_GROUP_VAL);
                    deinstall_very_first();
                }
                RegCloseKey(hKey);
            }
        }
        if(g_opt.drv_opt) {
            if(RegOpenKeyW(HKEY_LOCAL_MACHINE, NT_DBGPRNHK_REG_PATH, &hKey) != ERROR_SUCCESS) {
                print_log("Error: Can't store driver parameters in registry\n");
                exit(-4);
            }
            // write driver settings to registry
            #define DBGPRINT_DRV_OPT_RAW_WRITE_REG
            #include "dbgprint_opt_list.h"

    /*
            for(j=0; DbgPrintHook_Options[j].OptionName; j++) {
                if(!DbgPrintHook_Options[j].UpdateFlag)
                    continue;
                k = sizeof(ULONG);
                i = DbgPrintHook_Options[j].Value.OptDword;
                if(RegSetValueExW(hKey, DbgPrintHook_Options[j].OptionName, NULL, REG_DWORD, (PUCHAR)&i, k) != ERROR_SUCCESS) {
                    print_log("Error: Can't store driver parameters in registry\n");
                    exit(-4);
                }
            }
    */
            RegCloseKey(hKey);
        }
        if(g_opt.install_svc) {
            if(g_opt.svc_mode == SERVICE_DISABLED) {
                k = NtServiceStop(NT_DbgPrnHk_Client_SVC_NAME, 10);
                k = NtServiceRemove(NT_DbgPrnHk_Client_SVC_NAME);
                if(k == 1) {
                    print_log("DbgPrintLog service is uninstalled\n");
                }
            } else {
                // try to install (or re-install) DbgPrint capturing service
                k = NtServiceStop(NT_DbgPrnHk_Client_SVC_NAME, 10);
                k = NtServiceRemove(NT_DbgPrnHk_Client_SVC_NAME);
                k = NtServiceInstall(NT_DbgPrnHk_Client_SVC_NAME, PathToExec, FALSE, g_opt.svc_mode);
                if(k < 0) {
                    print_log("Error: Can't install DbgPrintLog service\n");
                    exit(-1);
                }
                // save command line in registry
                if(RegOpenKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\" NT_DbgPrnHk_Client_SVC_NAME, &hKey) != ERROR_SUCCESS) {
                    print_log("Error: Can't store parameters in registry\n");
                    exit(-4);
                }
                CmdLine2 = GetCommandLineW();
                if(!CmdLine2) {
                    exit(-1);
                }
                k = (wcslen(CmdLine2)+1)*sizeof(WCHAR);
                if(RegSetValueExW(hKey, NT_DbgPrnHk_Reg_CmdLineW, NULL, REG_SZ, (PUCHAR)CmdLine2, k) != ERROR_SUCCESS) {
                    print_log("Error: Can't store parameters in registry\n");
                    exit(-4);
                }
                k = (wcslen(g_opt.WDir)+1)*sizeof(WCHAR);
                if(RegSetValueExW(hKey, NT_DbgPrnHk_Reg_WDirW, NULL, REG_SZ, (PUCHAR)g_opt.WDir, k) != ERROR_SUCCESS) {
                    print_log("Error: Can't store parameters in registry\n");
                    exit(-4);
                }
                RegCloseKey(hKey);
            }
        }

        if(g_opt.install_drv || g_opt.install_svc || g_opt.drv_opt) {
            print_log("Changes will take effect after reboot or driver restart\n");
            print_log("Use 'net start/stop DbgPrintLog' to start/stop logging immediately\n");
            exit(0);
        }
    }
/*
    if(change_config) {
        print_log("Config changed\n");
        exit(0);
    }
*/

    if(!g_opt.input_driver) {
        // avoid multiple instances
        if(!g_opt.input_driver && AmIStarted()) {
            print_log("Multiple instances not allowed.\n");
            exit(-2);
        }

        // init lock for switching buffers
        hODS_lock = CreateMutex(NULL, FALSE, NULL);
        if(!hODS_lock || hODS_lock == (HANDLE)(-1)) {
            print_log("can't init mutex, (%#x)\n",
                GetLastError());
            exit(-1);
        }
    }
    if(sys_account) {
        // init as service
        if (!StartServiceCtrlDispatcher(dispatchTable)) {
            AddToMessageLog("StartServiceCtrlDispatcher failed.");
            exit(-1);
        }
    } else {
        // start keyboard event handling thread
        CreateThread(NULL, 0, get_kbd, NULL, 0, &thId);
        // get messages from user- and kernel-mode and store them to log-file
        get_messages();
    }

    exit(0);
} // end main()

